# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) Camptocamp SA
# Author: Arnaud Wüst
#
#
#    This file is part of the c2c_planning_management module
#    it contain classes used by c2c_planning_management/report/workload.py report and also by c2c_planning_management/hr_holidays.py
#    to do calculations on time periods
#
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#id)
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
from datetime import datetime, timedelta
import time
import pooler
import netsvc


class Days(object):
    """ a list of days """
    days_ids = []
    days = {}
    
    
    def __init__(self):
        """ constructor """
        self.days_ids = []
        self.days = {}
    
    def get_id(self, date):
        """ build an identifier from a date """
        return date.strftime("%Y-%m-%d")
    
    def append(self, day):
        """ add a day to the list """
        id = self.get_id(day.date_from)
        self.days_ids.append(id)
        self.days[id] = day
      
            
    def get_list(self):
        """ return the list of days contain in the object """
        result = []
        for id in self.days_ids:
            result.append(self.days[id])
        return result
      
        
    def get(self, date):
        """ return a day """
        id = self.get_id(date)
        return self.days[id]
      
      
    def update(self, day):              
        """ replace a day object by another in the list without changing his pos """
        id = self.get_id(day.date_from)
        self.days[id] = day
        
        
        
        
class Period(object):
    """ contain informations about periods""" 
    
    date_from = None
    date_to = None
    working_time = None #time to work between those two dates
    workload = None #amount of job planned between those two dates
    scale= None
    name = ""
    work_day = True  #does people work this period?
    project = []

    
    def __init__(self, date_from, date_to):
        """constructor """
        self.date_from = date_from
        self.date_to = date_to
        self.working_time = 0
        self.workload = 0
        self.work_day = True
        self.project = []
        
    def has_conflict(self):
        """ return true if this time unit has more planned work than working time """
        return self.working_time > self.workload
     
     
class Day(Period):   
        
    def __init__(self, date):  
        super(Day, self).__init__(date, date)
        self.scale = 'day'
        self.name =  date.strftime("%d.%m.%Y")
        

class Week(Period):
    def __init__(self, date_from, date_to):
        super(Week, self).__init__(date_from, date_to)
        self.scale = 'week'
        self.name = "Week %(week)s %(year)s" % {'week': date_from.isocalendar()[1], 'year': date_from.strftime("%Y") }
        

class Month(Period):
    def __init__(self, date_from, date_to):
        super(Month, self).__init__(date_from, date_to)
        self.scale = "Month"
        self.name = date_from.strftime("%m.%Y")
                


class User(object):            
    """contain informations about users """
    
    DEFAULT_WORKTIME = 42 # number of hours worked by week used for users not connected to employees 

    obj=  None
    
    lines = []
    _days_list = None
    boundary_start = None
    boundary_end = None
    
    working_week_time = None;
    
    def __init__(self, obj, boundary_start, boundary_end):
        """ constructor 
            obj is the orm object 
            boundary_* are date that define the limit of futur calculation. (given for optimization prupose)
        """
        self.obj= obj
        self.lines =  []
        self._days_list = None
        self.boundary_start = boundary_start
        self.boundary_end = boundary_end       
        
        self.working_week_time = None;
    
    def get_workload(self, cr, uid, period):
        """ return the workload for this user during the given period """
        
        days = self._get_days_list(cr, uid)
        result = 0
        #find inside the whole days list, the ones that fit in the period
        for d in days.get_list():
            if d.date_from >= period.date_from and d.date_to <= period.date_to:
                result += d.workload
        return result
    
        
    def get_working_time(self, cr, uid, period):
        """ return the working time for the given period """
        
        days = self._get_days_list(cr, uid)
        result = 0
        #find inside the whole days list, the ones that fit in the period
        for d in days.get_list(): 
            if d.date_from >= period.date_from and d.date_to <= period.date_to:
                result += d.working_time
        return result
        
        
    def get_week_working_time(self, cr, uid, user_id):
        """for a given user id, return the number of hours the user work per weeks """
        
        if self.working_week_time == None:
            
            self.working_week_time = self.DEFAULT_WORKTIME

            mypool = pooler.get_pool(cr.dbname)
            employee_obj = mypool.get('hr.employee')
            employee_id = employee_obj.search(cr, uid, [('user_id','=', user_id)])
            
            #if the user is linked to an employee
            if employee_id:
                employee = employee_obj.browse(cr, uid, employee_id[0])
                self.working_week_time = employee.regime
            else: 
                #no employee id
                netsvc.Logger().notifyChannel('report', netsvc.LOG_WARNING, "Report Error: There is no employee obj linked to the user id = %s. Assume he works %s hours per week" % (user_id, self.DEFAULT_WORKTIME))
        return float(self.working_week_time)
    
    
    
            
    def _get_days_list(self, cr, uid):
        """return a list of days between boundary dates that define working time and workload """
        
        if self._days_list == None:
            days = Days()

            
            ######################## 
            # first build the days list and define the workload for each days
            ########################
            
            current_date = self.boundary_start
            while current_date <= self.boundary_end:
    
                #TODO: improve this to take the holidays and the working hours in account to compute the working time. 
                # for now, we just assume the employees work from monday to friday, the same amount of time
                day = Day(current_date)
                if current_date.weekday() in [0,1,2,3,4]:
                    day.working_time = self.get_week_working_time(cr, uid, self.obj.id) / 5
                    day.work_day = True
                else:
                    day.working_time = 0
                    day.work_day = False
    
                days.append(day)
                current_date = current_date + timedelta(days=1)   
                
            ##########################
            # then spread the workload of lines over those days
            ##########################    
            
            for l in self.lines:
                #count the working day covered by this line and link project to days
                nbr_working_days = 0
                current_date = datetime(*(time.strptime(l.date_from, "%Y-%m-%d")[0:6]))
                days_to_update = []
                while current_date <= datetime(*(time.strptime(l.date_to, "%Y-%m-%d")[0:6])):
                    day = days.get(current_date)
                    if day.work_day:
                        nbr_working_days += 1 
                        #link the project to the day
                        proj =  (l.account_id.name_get(cr, uid, [l.account_id.id])[0][1],l.activity_id.name)
                        try: 
                            index = day.project.index(proj)
                            in_list = True
                        except Exception, e:
                            in_list = False
                        if not in_list: 
                            day.project.append(proj)     
                            
                        days_to_update.append(day)
                    current_date = current_date + timedelta(days=1)
       
                
                #compute the workload for each day
                if nbr_working_days > 0:
                    workload = l.unified_amount / nbr_working_days   
                else: 
                    workload = l.unified_amount
                
                #update days with workload
                for d in days_to_update:
                    d.project.sort()
                    d.workload += workload
                    days.update(d)

            self._days_list = days
        
        return self._days_list


        
    
    def get_projects(self, cr, uid, period):
        """ return the projects related to this user for this period """
        
        days = self._get_days_list(cr, uid)
        result = []
        #find inside the whole days list, the ones that fit in the period
        days_list = days.get_list()
        for d in days_list: 
            if d.date_from >= period.date_from and d.date_to <= period.date_to:
                for p in d.project:
                    try: 
                        index = result.index(p)
                        in_list = True
                    except Exception, e:
                        in_list = False
                    if not in_list:
                        result.append(p)
        return result
        
        
  
    def add_line(self, line):
        """add a line object (openerp object) to the user's lines list """        
        self.lines.append(line)

         